/*******************************************************************************
 (c) 2005-2014 Copyright, Real-Time Innovations, Inc.  All rights reserved.
 RTI grants Licensee a license to use, modify, compile, and create derivative
 works of the Software.  Licensee has the right to distribute object form only
 for use with RTI products.  The Software is provided "as is", with no warranty
 of any type, including any warranty for fitness for any purpose. RTI is under
 no obligation to maintain or support the Software.  RTI shall not be liable for
 any incidental or consequential damages arising out of the use or inability to
 use the software.
 ******************************************************************************/
/* profiles_publisher.c

   A publication of data of type profiles

   This file is derived from code automatically generated by the rtiddsgen
   command:

   rtiddsgen -language C -example <arch> profiles.idl

   Example publication of type profiles automatically generated by
   'rtiddsgen'. To test them follow these steps:

   (1) Compile this file and the example subscription.

   (2) Start the subscription with the command
       objs/<arch>/profiles_subscriber <domain_id> <sample_count>

   (3) Start the publication with the command
       objs/<arch>/profiles_publisher <domain_id> <sample_count>

   (4) [Optional] Specify the list of discovery initial peers and
       multicast receive addresses via an environment variable or a file
       (in the current working directory) called NDDS_DISCOVERY_PEERS.

   You can run any number of publishers and subscribers programs, and can
   add and remove them dynamically from the domain.


   Example:

       To run the example application on domain <domain_id>:

       On Unix:

       objs/<arch>/profiles_publisher <domain_id>
       objs/<arch>/profiles_subscriber <domain_id>

       On Windows:

       objs\<arch>\profiles_publisher <domain_id>
       objs\<arch>\profiles_subscriber <domain_id>


modification history
------------ -------
*/

#include "ndds/ndds_c.h"
#include "profiles.h"
#include "profilesSupport.h"
#include <stdio.h>
#include <stdlib.h>

#define NUM_PROFILE_FILES 1

/* Delete all entities */
static int publisher_shutdown(
        DDS_DomainParticipant *participant,
        struct DDS_DomainParticipantFactoryQos *factory_qos)
{
    DDS_ReturnCode_t retcode;
    int status = 0;

    retcode = DDS_DomainParticipantFactoryQos_finalize(factory_qos);
    if (retcode != DDS_RETCODE_OK) {
        printf("FactoryQos_finalize error %d\n", retcode);
        status = -1;
    }

    if (participant != NULL) {
        retcode = DDS_DomainParticipant_delete_contained_entities(participant);
        if (retcode != DDS_RETCODE_OK) {
            printf("delete_contained_entities error %d\n", retcode);
            status = -1;
        }

        retcode = DDS_DomainParticipantFactory_delete_participant(
                DDS_TheParticipantFactory,
                participant);
        if (retcode != DDS_RETCODE_OK) {
            printf("delete_participant error %d\n", retcode);
            status = -1;
        }
    }

    /* RTI Connext provides finalize_instance() method on
       domain participant factory for people who want to release memory used
       by the participant factory. Uncomment the following block of code for
       clean destruction of the singleton. */
    /*
        retcode = DDS_DomainParticipantFactory_finalize_instance();
        if (retcode != DDS_RETCODE_OK) {
            printf("finalize_instance error %d\n", retcode);
            status = -1;
        }
    */

    return status;
}

static int publisher_main(int domainId, int sample_count)
{
    DDS_DomainParticipant *participant = NULL;
    DDS_Publisher *publisher = NULL;
    DDS_Topic *topic = NULL;
    /* Volatile and transient local writers and profilesDataWriters */
    DDS_DataWriter *writer_volatile = NULL;
    DDS_DataWriter *writer_transient_local = NULL;
    profilesDataWriter *profiles_writer_volatile = NULL;
    profilesDataWriter *profiles_writer_transient_local = NULL;
    profiles *instance = NULL;
    DDS_ReturnCode_t retcode;
    DDS_InstanceHandle_t instance_handle = DDS_HANDLE_NIL;
    const char *type_name = NULL;
    int count = 0;
    /* Send a new sample every second */
    struct DDS_Duration_t send_period = { 1, 0 };

    struct DDS_DomainParticipantFactoryQos factory_qos =
            DDS_DomainParticipantFactoryQos_INITIALIZER;
    static const char *myUrlProfile[NUM_PROFILE_FILES] = {
        "file://./my_custom_qos_profiles.xml"
    };

    /* There are several different approaches for loading QoS profiles from XML
     * files (see Configuring QoS with XML chapter in the RTI Connext Core
     * Libraries and Utilities User's Manual). In this example we illustrate
     * two of them:
     *
     * 1) Creating a file named USER_QOS_PROFILES.xml, which is loaded,
     * automatically by the DomainParticipantFactory. In this case, the file
     * defines a QoS profile named volatile_profile that configures reliable,
     * volatile DataWriters and DataReaders.
     *
     * 2) Adding XML documents to the DomainParticipantFactory using its
     * Profile QoSPolicy (DDS Extension). In this case, we add
     * my_custom_qos_profiles.xml to the url_profile sequence, which stores
     * the URLs of all the XML documents with QoS policies that are loaded by
     * the DomainParticipantFactory aside from the ones that are automatically
     * loaded.
     * my_custom_qos_profiles.xml defines a QoS profile named
     * transient_local_profile that configures reliable, transient local
     * DataWriters and DataReaders.
     */

    /* To load my_custom_qos_profiles.xml, as explained above, we need to modify
     * the  DDSTheParticipantFactory Profile QoSPolicy */
    DDS_DomainParticipantFactory_get_qos(
            DDS_TheParticipantFactory,
            &factory_qos);

    /* We are only going to add one XML file to the url_profile sequence, so our
     * NUM_PROFILE_FILES is 1 (defined at start) */
    DDS_StringSeq_from_array(
            &(factory_qos.profile.url_profile),
            (const char **) myUrlProfile,
            NUM_PROFILE_FILES);

    /* The XML file will be loaded from the working directory. That means, you
     * need to run the example like this:
     * ./objs/<architecture>/profiles_publisher
     * (see README.txt for more information on how to run the example).
     *
     * Note that you can specify the absolute path of the XML QoS file to avoid
     * this problem.
     */
    retcode = DDS_DomainParticipantFactory_set_qos(
            DDS_TheParticipantFactory,
            &factory_qos);
    if (retcode != DDS_RETCODE_OK) {
        printf("set_qos error %d\n", retcode);
        publisher_shutdown(participant, &factory_qos);
        return -1;
    }

    /* Our default Qos profile, volatile_profile, sets the participant name.
     * This is the only participant_qos policy that we change in our
     * example. As this is done in the default QoS profile, we don't need
     * to specify its name, so we can create the participant using the
     * create_participant() method rather than using
     * create_participant_with_profile().  */
    participant = DDS_DomainParticipantFactory_create_participant(
            DDS_TheParticipantFactory,
            domainId,
            &DDS_PARTICIPANT_QOS_DEFAULT,
            NULL /* listener */,
            DDS_STATUS_MASK_NONE);
    if (participant == NULL) {
        printf("create_participant error\n");
        publisher_shutdown(participant, &factory_qos);
        return -1;
    }

    /* We haven't changed the publisher_qos in any of QoS profiles we use in
     * this example, so we can just use the create_publisher() method. If you
     * want to load an specific profile in which you may have changed the
     * publisher_qos, use the create_publisher_with_profile() method. */
    publisher = DDS_DomainParticipant_create_publisher(
            participant,
            &DDS_PUBLISHER_QOS_DEFAULT,
            NULL /* listener */,
            DDS_STATUS_MASK_NONE);
    if (publisher == NULL) {
        printf("create_publisher error\n");
        publisher_shutdown(participant, &factory_qos);
        return -1;
    }

    /* Register type before creating topic */
    type_name = profilesTypeSupport_get_type_name();
    retcode = profilesTypeSupport_register_type(participant, type_name);
    if (retcode != DDS_RETCODE_OK) {
        printf("register_type error %d\n", retcode);
        publisher_shutdown(participant, &factory_qos);
        return -1;
    }

    /* We haven't changed the topic_qos in any of QoS profiles we use in this
     * example, so we can just use the create_topic() method. If you want to
     * load an specific profile in which you may have changed the topic_qos,
     * use the create_topic_with_profile() method. */
    topic = DDS_DomainParticipant_create_topic(
            participant,
            "Example profiles",
            type_name,
            &DDS_TOPIC_QOS_DEFAULT,
            NULL /* listener */,
            DDS_STATUS_MASK_NONE);
    if (topic == NULL) {
        printf("create_topic error\n");
        publisher_shutdown(participant, &factory_qos);
        return -1;
    }

    /* Volatile writer -- As volatile_profile is the default qos profile
     * we don't need to specify the profile we are going to use, we can
     * just call create_datawriter passing DDS_DATAWRITER_QOS_DEFAULT. */
    writer_volatile = DDS_Publisher_create_datawriter(
            publisher,
            topic,
            &DDS_DATAWRITER_QOS_DEFAULT, /* Default datawriter_qos */
            NULL /* listener */,
            DDS_STATUS_MASK_NONE);
    if (writer_volatile == NULL) {
        printf("create_datawriter error\n");
        publisher_shutdown(participant, &factory_qos);
        return -1;
    }

    /* Transient Local writer -- In this case we use
     * create_datawriter_with_profile, because we have to use a profile other
     * than the default one. This profile has been defined in
     * my_custom_qos_profiles.xml, but since we already loaded the XML file
     * we don't need to specify anything else. */
    writer_transient_local = DDS_Publisher_create_datawriter_with_profile(
            publisher,                 /* publisher which creates the writer */
            topic,                     /* DDS_topic */
            "profiles_Library",        /* library_name */
            "transient_local_profile", /*profile_name */
            NULL /* listener */,
            DDS_STATUS_MASK_NONE);
    if (writer_transient_local == NULL) {
        printf("create_datawriter_with_profile error\n");
        publisher_shutdown(participant, &factory_qos);
        return -1;
    }

    profiles_writer_volatile = profilesDataWriter_narrow(writer_volatile);
    if (profiles_writer_volatile == NULL) {
        printf("DataWriter narrow error\n");
        publisher_shutdown(participant, &factory_qos);
        return -1;
    }

    profiles_writer_transient_local =
            profilesDataWriter_narrow(writer_transient_local);
    if (profiles_writer_transient_local == NULL) {
        printf("DataWriter narrow error\n");
        publisher_shutdown(participant, &factory_qos);
        return -1;
    }

    /* Create data sample for writing */

    instance = profilesTypeSupport_create_data_ex(DDS_BOOLEAN_TRUE);

    if (instance == NULL) {
        printf("profilesTypeSupport_create_data error\n");
        publisher_shutdown(participant, &factory_qos);
        return -1;
    }

    /* For a data type that has a key, if the same instance is going to be
       written multiple times, initialize the key here
       and register the keyed instance prior to writing */
    /*
        instance_handle = profilesDataWriter_register_instance(
            profiles_writer, instance);
    */

    /* Main loop */
    for (count = 0; (sample_count == 0) || (count < sample_count); ++count) {
        printf("Writing profiles, count %d\n", count);

        /* Modify the data to be written here */
        strcpy(instance->profile_name, "volatile_profile");
        instance->x = count;

        printf("Writing profile_name = %s,\t x = %d\n",
               instance->profile_name,
               instance->x);

        retcode = profilesDataWriter_write(
                profiles_writer_volatile,
                instance,
                &instance_handle);
        if (retcode != DDS_RETCODE_OK) {
            printf("write error %d\n", retcode);
        }

        strcpy(instance->profile_name, "transient_local_profile");
        instance->x = count;

        printf("Writing profile_name = %s,\t x = %d\n\n",
               instance->profile_name,
               instance->x);

        retcode = profilesDataWriter_write(
                profiles_writer_transient_local,
                instance,
                &instance_handle);
        if (retcode != DDS_RETCODE_OK) {
            printf("write error %d\n", retcode);
        }

        NDDS_Utility_sleep(&send_period);
    }

    /*
        retcode = profilesDataWriter_unregister_instance(
            profiles_writer, instance, &instance_handle);
        if (retcode != DDS_RETCODE_OK) {
            printf("unregister instance error %d\n", retcode);
        }
    */

    /* Delete data sample */
    retcode = profilesTypeSupport_delete_data_ex(instance, DDS_BOOLEAN_TRUE);
    if (retcode != DDS_RETCODE_OK) {
        printf("profilesTypeSupport_delete_data error %d\n", retcode);
    }

    /* Cleanup and delete delete all entities */
    return publisher_shutdown(participant, &factory_qos);
}

#if defined(RTI_WINCE)
int wmain(int argc, wchar_t **argv)
{
    int domainId = 0;
    int sample_count = 0; /* infinite loop */

    if (argc >= 2) {
        domainId = _wtoi(argv[1]);
    }
    if (argc >= 3) {
        sample_count = _wtoi(argv[2]);
    }

    /* Uncomment this to turn on additional logging
    NDDS_Config_Logger_set_verbosity_by_category(
        NDDS_Config_Logger_get_instance(),
        NDDS_CONFIG_LOG_CATEGORY_API,
        NDDS_CONFIG_LOG_VERBOSITY_STATUS_ALL);
    */

    return publisher_main(domainId, sample_count);
}
#elif !(defined(RTI_VXWORKS) && !defined(__RTP__)) && !defined(RTI_PSOS)
int main(int argc, char *argv[])
{
    int domainId = 0;
    int sample_count = 0; /* infinite loop */

    if (argc >= 2) {
        domainId = atoi(argv[1]);
    }
    if (argc >= 3) {
        sample_count = atoi(argv[2]);
    }

    /* Uncomment this to turn on additional logging
    NDDS_Config_Logger_set_verbosity_by_category(
        NDDS_Config_Logger_get_instance(),
        NDDS_CONFIG_LOG_CATEGORY_API,
        NDDS_CONFIG_LOG_VERBOSITY_STATUS_ALL);
    */

    return publisher_main(domainId, sample_count);
}
#endif

#ifdef RTI_VX653
const unsigned char *__ctype = NULL;

void usrAppInit()
{
    #ifdef USER_APPL_INIT
    USER_APPL_INIT; /* for backwards compatibility */
    #endif

    /* add application specific code here */
    taskSpawn(
            "pub",
            RTI_OSAPI_THREAD_PRIORITY_NORMAL,
            0x8,
            0x150000,
            (FUNCPTR) publisher_main,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0);
}
#endif
